package pacworld;

import agent.*;
import java.io.*;
import java.util.Vector;
import java.awt.Point;
import java.lang.Math;

/** Represents a state in the package world. */
public class PackageState extends State {

  /** Construct a new package state. */
  public PackageState() {
  }
  
    /** Array containing state information for the agents. */
    protected PacAgentRep[] agentReps;

    protected Vector packages;
    protected Location[] destinations;
    protected String[] messages;

    protected int numAgents = 0;
    protected int numPackages = 0;
    protected int origNumPackages = 0;
    protected int mapSize = 0;
    protected int numDests = 0;

    protected int idleCount = 0;

    /** The number of actions not including idling. */
    protected int workCount = 0;

    /** The number of moves with a package in hand. */
    protected int pacCount = 0;

    protected int messageCount = 0;
    protected int totalMessageLength = 0;
    protected int nonProductiveCount = 0;
    protected PacGUI gui = null;

  /** An array to that contains the locations of objects in the world. */
  protected Object[][] map;
    

  /** Return the default initial state for the vacuum world. */
  public static PackageState getInitState(Vector agts, int numPackages, 
					  int numDests, int mapSize) {
    PackageState state;
    int i, x, y;
    Location loc;

    state = new PackageState();
    state.origNumPackages = state.numPackages = numPackages;
    state.numAgents = agts.size();
    state.agentReps = new PacAgentRep[state.numAgents];
    state.packages = new Vector();

    state.mapSize = mapSize;
    state.map = new Object[mapSize][mapSize];
    for (i = 0; i < mapSize; i++)
      for (int j = 0; j < mapSize; j++)
        state.map[i][j] = null;

    state.numDests = numDests;
    state.destinations = new Location[numDests];
    // Note: Math.random() generates numbers s.t. 0.0 <= n < 1.0
    // Casting a float to an int, truncates the fractional part
    // we don't want to put desitnations on the edge of the map
    for (i = 0; i < numDests; i++) {
      x = (int)(Math.random() * (mapSize - 2)) + 1;
      y = (int)(Math.random() * (mapSize - 2)) + 1;
      state.destinations[i] = new Location(x, y);
    }

    for (i = 0; i < state.numPackages; i++) {
      Package p = new Package(i);
      int destId;
      do {
        x = (int)(Math.random() * mapSize);
        y = (int)(Math.random() * mapSize);
	destId = (int)(Math.random() * numDests);
      } while (state.map[x][y] != null || 
	       (state.destinations[destId].getX() == x &&
		state.destinations[destId].getY() == y));     
      p.setX(x);
      p.setY(y);
      p.setDestination(destId, state.destinations[destId].getX(),
		       state.destinations[destId].getY());
      state.packages.add(p);
      state.map[x][y] = p;
    }

    for (i = 0; i < state.numAgents; i++) {
      PacAgentRep arep = (PacAgentRep)(agts.elementAt(i));
      do {
        x = (int)(Math.random() * mapSize);
        y = (int)(Math.random() * mapSize);
      } while (state.map[x][y] != null);
      arep.setX(x);
      arep.setY(y);
      state.agentReps[i] = arep;
      state.map[x][y] = arep;
    }

    state.messages = new String[state.numAgents];
    for (i = 0; i < state.numAgents; i++)
      state.messages[i] = null;
    
    // state.printState();

    return state;
  }

  public void printState() {
    int i;
    System.out.println("Agents:");
    for (i = 0; i < agentReps.length; i++) {
      PacAgentRep arep = agentReps[i];
      System.out.println("Agent " + i + ": x = " + arep.getX() + " y = " + arep.getY());
    }

    System.out.println("Packages:");
    for (i = 0; i < packages.size(); i++) {
      Package p = (Package)packages.elementAt(i);
      if (p != null)
        System.out.println("Package " + i + ": x = " + p.getX() + " y = " + p.getY() + " destination = (" + p.getDestX() + "," + p.getDestY() + ")");
      else
	System.out.println("Package " + i + ": Delivered Successfully");
    }

    System.out.println("Destinations:");
    for (i = 0; i < numDests; i++) {
      Location p = destinations[i];
      System.out.println("Destination " + i + ": x = " + p.getX() + " y = " + p.getY());
    }
  }

  /** Modify the state so that the agent is holding the selected package.
      If the package is already held by an agent, then do nothing. */
  public void pickup(PacAgentRep a, Package p) {
    // technically, the following fields should be incremented
    // by the Pickup action (so that trying to pick up something other
    // than a package counts as work!)
    workCount++;
    nonProductiveCount++;

    if (p.getAgent() != null) 
      System.out.println("Agent " + a.getId() + " tried to grab a " +
			 "package held by another agent!");
    else {
      a.setHeldPackage(p);
      p.setAgent(a);
    }
  }

  /** Change the agent's position. Update both the list of agent locations
      and the map.*/
  public void repositionAgent(PacAgentRep a, int newX, int newY) {
    int oldx, oldy;

    workCount++;
    nonProductiveCount++;
    oldx = a.getX();
    oldy = a.getY();
    a.setX(newX);
    a.setY(newY);
    // make sure we haven't moved something to agent's position first
    if (map[oldx][oldy] == a)
      map[oldx][oldy] = null;      
    map[newX][newY] = a;
  }

  /** Change a package's position. Update the package's location and
      the map. */
  public void repositionPackage(Package p, int newX, int newY) {
    int oldx, oldy;

    pacCount++;
    oldx = p.getX();
    oldy = p.getY();
    p.setX(newX);
    p.setY(newY);
    // it is possible that the agent have moved in the direction of
    // of the package, so we must be careful not to overwrite it.
    if (map[oldx][oldy] instanceof Package)
      map[oldx][oldy] = null;      
    map[newX][newY] = p;
  }


  /** Set an indicator that the robot bumped into an obstacle 
      during its previous move. */
  public void setBump(PacAgentRep a, boolean bump) {
    a.setBumped(bump);
  }

  /** Return true if the specified location has an obstacle in it. */
  public boolean hasObstacle(int x, int y) {

    if (map[x][y] != null)
      return true;
    else
      return false;
  }

  /** Return the object located at the specified location. If there is
      no object there, returns null. */
  public Object getObjectAt(int x, int y) {
    return map[x][y];
  }

  /** Return true if the agent bumped into an obstacle on its previous
      move. */
  public boolean bumped(PacAgentRep a) {
    return a.hasBumped();
  }

  /** Return true if the location is within bounds of the state's map. */
  public boolean inBounds(int x, int y) {

    if (x >= 0 && x < mapSize && y >= 0 && y < mapSize)
      return true;
    else
      return false;
  }

  public void dropPackage(PacAgentRep a, int dropX, int dropY) {
    // technically, the following fields should be incremented
    // by the Dropoff action (so that trying to drop something in
    // an occupied square counts as work!)
    workCount++;
    nonProductiveCount++;
    Package p = a.getHeldPackage();
    if (p != null) {
      p.setAgent(null);
      a.setHeldPackage(null);
      int x = p.getX();
      int y = p.getY();

      // if package is dropped at its destination, then it disappears
      if (dropX == p.getDestX() && dropY == p.getDestY()) {
	packages.set(p.getId(), null);
	p = null;
	map[x][y] = null;
	numPackages--;
	nonProductiveCount = 0;
      }
      // if package is dropped in a different location from where it was
      else if (x != dropX || y != dropY) {
        map[x][y] = null;
	map[dropX][dropY] = p;
	p.setX(dropX);
	p.setY(dropY);
      }
      // otherwise, package stays where it is
    }
    else
      System.out.println("Not holding a package.");
  }

  public int getNumPackages() {
    return numPackages;
  }

  public int getOrigNumPackages() {
    return origNumPackages;
  }

  public int getMapSize() {
    return mapSize;
  }

  public int getIdleCount() {
    return idleCount;
  }

  public void setIdleCount(int ic) {
    if (ic > idleCount)
      nonProductiveCount++;
    idleCount = ic;
  }

  public Vector getPackages() {
    return packages;
  }

  public String[] getMessages() {
    return messages;
  }

  public Location[] getDestinations() {
    return destinations;
  }

  public int getWorkCount() {
    return workCount;
  }

  public int getPacCount() {
    return pacCount;
  }

  public int getMessageCount() {
    return messageCount;
  }

  public int getTotalMessageLength() {      
    return totalMessageLength;
  }

  public int getNonProductiveCount() {
    return nonProductiveCount;
  }

  /** Set whatever message was said by the agent in its previous move. */
  public void setMessage(PacAgentRep a, String s) {
    messages[a.getId()] = s;
    if (s != null) {
      workCount++;
      messageCount++;
      totalMessageLength += s.length();
    }
  }

  /** Print an output of the state to the screen. */
  public void display() {
    if (gui != null)
      gui.drawMap(this);
  }

  public void setGUI(PacGUI pg) {
    gui = pg;
  }

  /** Returns the map. This is needed for the CustomPanel class. */
  public Object[][] getMap() {
    return map;
  }

  public PacAgentRep[] getAgentReps() {
    return agentReps;
  }

}



